#include "buffer_manager.hpp"

#include "exceptions.hpp"
#include "vulkan_allocator.hpp"
#include "vulkan_instance.hpp"
#include <gz-util/string/to_string.hpp>

namespace gz::vlk {
	std::string BufferInfo::toString() const {
		return "<index: " + gz::toString(index) + ", offset: " + gz::toString(offset) + ", count:" + gz::toString(count) + ">";
	}

	template class BufferManager<Vertex2D, uint32_t>;
	template class BufferManager<Vertex3D, uint32_t>;
	template class BufferManager<Vertex2D, uint16_t>;
	template class BufferManager<Vertex3D, uint16_t>;
	template<VertexType VertexT, SupportedIndexType IndexT>
	BufferManager<VertexT, IndexT>::BufferManager(VulkanInstance& instance) 
		: vk(instance)
	{
		bLog = vk.createSublog(true, "BufferManager", gz::Color::BO_CYAN);
		vk.registerCleanupCallback(std::bind(&BufferManager::cleanup, this));
	}



	template<VertexType VertexT, SupportedIndexType IndexT>
	vk::Buffer BufferManager<VertexT, IndexT>::getVertexBuffer(const BufferInfo& vertexBufferInfo)const {
		if (vertexBuffers.size() <= vertexBufferInfo.index) {
			throw VkUserError("Invalid buffer index: " + gz::toString(vertexBufferInfo.index), "BufferManager::getVertexBuffer");
		}
		return vertexBuffers.at(vertexBufferInfo.index).buffer;
	}
	template<VertexType VertexT, SupportedIndexType IndexT>
	vk::Buffer BufferManager<VertexT, IndexT>::getIndexBuffer(const BufferInfo& indexBufferInfo)const {
		if (indexBuffers.size() <= indexBufferInfo.index) {
			throw VkUserError("Invalid buffer index: " + gz::toString(indexBufferInfo.index), "BufferManager::getIndexBuffer");
		}
		return indexBuffers.at(indexBufferInfo.index).buffer;
	}


	// helpers
	bool searchForSpaceInBuffers(std::vector<Buffer>& buffers, uint32_t allocSize, BufferInfo& bufferInfo) {
		// search for allocSize space in buffers, return true and set bufferinfo to found space if found
		for (auto bufferIt = buffers.begin(); bufferIt != buffers.end(); bufferIt++) {
			try {
				bufferInfo.offset = bufferIt->allocator.allocate(allocSize, 1);
				bufferInfo.index = bufferIt - buffers.begin();
				return true;
			}
			catch (...) {}
		}
		return false;
	};
	template<typename T>
	void allocateNewBuffer(std::vector<Buffer>& buffers, uint32_t count, VulkanInstance& vk) {
		uint32_t bufferSize = sizeof(T) * count;
		buffers.emplace_back(Buffer{.buffer = vk::Buffer(), .bufferMemory = MemoryInfo(), .bufferSize = bufferSize, .allocator = DynBlockAllocator(bufferSize)});
		if constexpr (VertexType<T>) {
			vk.createVertexBuffer<T>(count, buffers.back().buffer, buffers.back().bufferMemory, buffers.back().bufferSize);
		}
		else if constexpr (SupportedIndexType<T>) {
			vk.createIndexBuffer<T>(count, buffers.back().buffer, buffers.back().bufferMemory, buffers.back().bufferSize);
		}
		else {
			static_assert(VertexType<T> or SupportedIndexType<T>, "T is not vertex or index type");
		}
	};


	template<VertexType VertexT, SupportedIndexType IndexT>
	template<util::ContiguousRange<VertexT> VertexRange, util::ContiguousRange<IndexT> IndexRange>
	std::pair<BufferInfo, BufferInfo> BufferManager<VertexT, IndexT>::addVertices(const VertexRange& vertices, const IndexRange& indices) {
		std::pair<BufferInfo, BufferInfo> bufferInfos;
		bufferInfos.first.count = vertices.size();
		// search for space for vertex buffer
		if (!searchForSpaceInBuffers(vertexBuffers, sizeof(VertexT) * vertices.size(), bufferInfos.first)) {
			bLog.log0("Allocating new vertex buffer nr.", vertexBuffers.size());
			// allocate new vertex buffer 
			allocateNewBuffer<VertexT>(vertexBuffers, std::max(vertices.size(), TODO_VERTICES_COUNT), vk);
			bufferInfos.first.index = vertexBuffers.size() - 1;
			bufferInfos.first.offset = vertexBuffers.back().allocator.allocate(sizeof(VertexT) * vertices.size(), 1);
		}
		// copy vertices to vertex buffer
		Buffer& vertexBuffer = vertexBuffers.at(bufferInfos.first.index);
		vk.copyToDeviceBuffer(vertices.data(), sizeof(VertexT) * vertices.size(), vertexBuffer.buffer, vertexBuffer.bufferMemory, bufferInfos.first.offset);

		bufferInfos.second.count = indices.size();
		// search for space for index buffer
		if (!searchForSpaceInBuffers(indexBuffers, sizeof(IndexT) * indices.size(), bufferInfos.second)) {
			bLog.log0("Allocating new index buffer nr.", indexBuffers.size());
			// allocate new index buffer 
			allocateNewBuffer<IndexT>(indexBuffers, std::max(indices.size(), TODO_INDICES_COUNT), vk);
			bufferInfos.second.index = indexBuffers.size() - 1;
			bufferInfos.second.offset = indexBuffers.back().allocator.allocate(sizeof(IndexT) * indices.size(), 1);
		}
		// copy indices to index buffer
		Buffer& indexBuffer = indexBuffers.at(bufferInfos.first.index);
		vk.copyToDeviceBuffer(indices.data(), sizeof(IndexT) * indices.size(), indexBuffer.buffer, indexBuffer.bufferMemory, bufferInfos.second.offset);
		bLog.log0("addVertices: VertexBufferInfo:", bufferInfos.first, "IndexBufferInfo", bufferInfos.second);
		return bufferInfos;
	}
	template std::pair<BufferInfo, BufferInfo> BufferManager<Vertex2D, uint32_t>::addVertices(const std::vector<Vertex2D>& vertices, const std::vector<uint32_t>& indices);

	
	template<VertexType VertexT, SupportedIndexType IndexT>
	void BufferManager<VertexT, IndexT>::removeVertices(BufferInfo& vertexBufferInfo, BufferInfo& indexBufferInfo) {
		if (vertexBuffers.size() <= vertexBufferInfo.index) {
			throw VkUserError("Invalid vertex buffer index:" + gz::toString(vertexBufferInfo.index), "BufferManager::removeVertices");
		}
		Buffer& vertexBuffer = vertexBuffers.at(vertexBufferInfo.index);
		vertexBuffer.allocator.free(vertexBufferInfo.offset);
		vertexBufferInfo.index = BUFFER_NOT_INITIALIZED;
		vertexBufferInfo.offset = BUFFER_NOT_INITIALIZED;
		vertexBufferInfo.count = BUFFER_NOT_INITIALIZED;

		if (indexBuffers.size() <= indexBufferInfo.index) {
			throw VkUserError("Invalid index buffer index:" + gz::toString(indexBufferInfo.index), "BufferManager::removeVertices");
		}
		Buffer& indexBuffer = indexBuffers.at(indexBufferInfo.index);
		indexBuffer.allocator.free(indexBufferInfo.offset);
		indexBufferInfo.index = BUFFER_NOT_INITIALIZED;
		indexBufferInfo.offset = BUFFER_NOT_INITIALIZED;
		indexBufferInfo.count = BUFFER_NOT_INITIALIZED;
	}


	template<VertexType VertexT, SupportedIndexType IndexT>
	void BufferManager<VertexT, IndexT>::cleanup() {
		for (auto bufferIt = vertexBuffers.begin(); bufferIt != vertexBuffers.end(); bufferIt++) {
			// check if all blocks are free
			for (auto blockIt = bufferIt->allocator.begin(); blockIt != bufferIt->allocator.end(); blockIt++) {
				if (blockIt->free) { continue; }
				bLog.warning("cleanup: vertex buffer was not removed:", *blockIt, "index:", bufferIt - vertexBuffers.begin());
			}
			vk.destroyBuffer(bufferIt->buffer, bufferIt->bufferMemory);
		}
		vertexBuffers.clear();
		for (auto bufferIt = indexBuffers.begin(); bufferIt != indexBuffers.end(); bufferIt++) {
			// check if all blocks are free
			for (auto blockIt = bufferIt->allocator.begin(); blockIt != bufferIt->allocator.end(); blockIt++) {
				if (blockIt->free) { continue; }
				bLog.warning("cleanup: index buffer was not removed:", *blockIt, "index:", bufferIt - indexBuffers.begin());
			}
			vk.destroyBuffer(bufferIt->buffer, bufferIt->bufferMemory);
		}
		indexBuffers.clear();
	}
	
	
}  // namespace gz::vlk
