#include "dynamic_size_block_allocator.hpp"

#include "exceptions.hpp"
#include "vulkan_settings.hpp"

#include <gz-util/string/conversion.hpp>

namespace gz::vlk {
	std::string Block::toString() const {
		return std::string("<offset: " + gz::toHexString(offset, 0) + ", size: " + gz::toHexString(size, 0) + ", free: " + gz::toString(free) + ">");
	}


	DynBlockAllocator::DynBlockAllocator(uint32_t size) {
        /* LogCreateInfo logCI { */
        /* 	.logfile = "DynBlockAllocator.log", */
        /* 	.storeLog = false, */
        /* 	.prefix = "BlockAllocator", */
        /* 	.prefixColor = Color::LI_GREEN, */
        /* 	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR, */
		/* }; */
        /* aLog = Log(std::move(logCI)); */
		blocks.emplace_front(Block{ .size = size, .offset = 0, .free = true });
	}

	uint32_t DynBlockAllocator::allocate(uint32_t size, uint32_t alignment) {
		// go through blocks in memory
		for (auto block = blocks.begin(); block != blocks.end(); block++) {
			if (!block->free) { continue; }
			if (block->size >= size) {
				// check if alignment is ok
				if (block->offset % alignment != 0) {
					// move non-aligned space to previous block
					// new offset === offset + (align - offset % align)
					size_t moveToPreviousBlock = alignment - block->offset % alignment;
					// check if still large enough
					if (block->size - moveToPreviousBlock < size) { continue; }
					block--;
					block->size += moveToPreviousBlock;
					block++;
					block->offset += moveToPreviousBlock;
					block->size -= moveToPreviousBlock;
				}
				// if the block is larger than the needed size, split the block
				if (block->size > size) {
					// emplace free part of block after used part of block
					auto newBlock = ++block;
					block--;
					blocks.emplace(newBlock,
							Block{ block->size - size, block->offset + size, true });
					/* block--; */
					block->size = size;
				}
				block->free = false;
				/* aLog.log0("allocate: (memory", memoryInfo.memoryTypeIndex, "-", gz::toString(deviceMemory - memory[memoryInfo.memoryTypeIndex].begin()) + ") Blocks:", deviceMemory->blocks); */
				return block->offset;
			}
		}			
		// if no space, throw exception
		throw VkException("Out of memory", "DynBlockAllocator::allocate");
	}


	void DynBlockAllocator::free(uint32_t offset) {
		// go through blocks in memory
		for (auto block = blocks.begin(); block != blocks.end(); block++) {
			if (block->offset != offset) { continue; }
			block->free = true;
			if (block != blocks.begin()) {
				// merge with previous block, if free
				auto otherBlock = block;
				otherBlock--;
				if (otherBlock->free) {
					otherBlock->size += block->size;
					blocks.erase(block);
					block = otherBlock;
				}
			}
			if (block != --(blocks.end())) {
				// merge with next block, if free
				auto otherBlock = block;
				otherBlock++;
				if (otherBlock->free) {
					block->size += otherBlock->size;
					blocks.erase(otherBlock);
				}
			}
			return;
		}
		throw VkUserError("Trying to free block that was not allocated from this DynBlockAllocator", "DynBlockAllocator::free");
	}
}  // namespace gz::vlk
