#pragma once

#include "renderer.hpp"

#include "shape.hpp"
#include "vulkan_allocator.hpp"
#include "vulkan_util.hpp"

namespace gz::vlk {
    enum Pipelines2D {
        R2Render,
    };

    /**
     * @brief 2D Renderer
     * @details
     *  @section r3d_options Options
     *   - `RENDERER_2D_RENDER_TO_OWN_IMAGES`: render to own images
     *     - `RENDERER_2D_COPY_TO_SWAP_CHAIN`: copy the image to the swap chain after rendering it to the own image (swap chain images have to be created with TRANSFER_DST_BIT)
     *     - `RENDERER_2D_SAMPLE_FROM_RENDER_IMAGES`: enable sampling from the rendered images
     *   - `RENDERER_2D_RENDER_TO_SWAP_CHAIN`: render directly onto the swap chain
     */
    class Renderer2D : public Renderer {
        public:
            /**
             * @brief Create a 2D renderer
             * @details
             *  -# @ref VulkanInstance::registerCleanupCallback "register" @ref cleanup() "cleanup callback"
             *  -# @ref VulkanInstance::registerSwapChainRecreateCallback "register" @ref swapChainRecreateCallback "swapChain recreation callback"
             *  -# create command buffers
             *  -# create vertex & index buffers
             *  -# call createDescriptorResources (if `RENDERER_2D_SAMPLE_FROM_RENDER_IMAGES`)
             *  -# call initSwapChainDependantResources
             */
            Renderer2D(VulkanInstance& instance, TextureManager& textureManager);
            /**
             * @name Rendering
             */
            /// @{
            /**
             * @brief Begin rendering a new frame
             * @details
             *  As of now, this does (in the same command buffer from graphicsPool)
             *  -# begin render pass 
             *      - image layout: vk::ImageLayout::eColorAttachmentOptimal
             *      - clear image
             */
            void beginFrameDraw(uint32_t imageIndex, uint32_t currentFrame);
            /**
             * @brief End rendering a frame
             *  -# end render pass 
             *      - image layout: vk::ImageLayout::eTransferSrcOptimal
             *  -# copy image to swapChain image with same imageIndex (if RENDERER_2D_COPY_TO_SWAP_CHAIN)
             */
            void endFrameDraw(uint32_t imageIndex, uint32_t currentFrame);

            void drawShape(Shape* shape, uint32_t imageIndex, uint32_t currentFrame);
            void initializeShape(Shape* shape);
            /**
             * @brief Copies the vertices from shapes into the vertex buffer, using a staging buffer
             */
            /* void fillVertexBufferWithShapes(); */
            /* void fillIndexBufferWithShapes(); */
            /// @}

        private:
            /**
             * @brief Destroy all vulkan objects owned by this object
             * @details:
             *  Does:
             *  -# destroy descriptor layout and pool (if `RENDERER_2D_SAMPLE_FROM_RENDER_IMAGES`)
             *  -# call cleanupSwapChainDependantResources()
             *  -# call Renderer::cleanup_()
             */
            void cleanup();

            /* std::vector<Shape> shapes; */
            /* size_t shapesVerticesCount = 0; */
            /* size_t shapesIndicesCount = 0; */
            BufferManager<Vertex2D, uint32_t> shapeBufferManager;

            /* void recordCommandBuffer(uint32_t imageIndex, uint32_t currentFrame); */

            /**
             * @brief Create the render image resources
             * @details
             *  If `RENDERER_2D_RENDER_TO_SWAP_CHAIN`:
             *  - renderImages are null handles
             *  - renderImageMemory are null handles
             *  - renderImageViews are views to VulkanInstance::getScImages()
             *  - renderImageSampler is null handle
             *
             *  Else:
             *  - renderImages are images on renderImageMemory
             *  - renderImageViews are views to renderImages
             *  - renderImageSampler is a combined image sampler
             */
            void createImageResources();
            /// @}

            /**
             * @name Render pass
             * @details
             *  Attachments:
             *  - color blend:
             *      - loadOp = vk::AttachmentLoadOp::eClear
             *      - storeOp = vk::AttachmentStoreOp::eStore
             *      - initialLayout = vk::ImageLayout::eColorAttachmentOptimal
             *      - finalLayout = vk::ImageLayout::eColorAttachmentOptimal
             *      - stencil load/store = dont care
             */
            /// @{
            /**
             * @brief Create a render pass
             */
            void createRenderPass();
            vk::RenderPass renderPass;
            /// @}

            /**
             * @brief Framebuffer for renderImageViews
             */
            std::vector<vk::Framebuffer> framebuffers;

            // PIPELINE
            PipelineContainer<Pipelines2D> pipelines;

        /**
         * @name Desciptors
         * @details 
         *  
         *  @section r2d_desc Descriptors
         *   Renderer2D has one descriptor set.
         *   @subsection r2d_desc1 Set 1: For sampling from the rendered image 
         *    If `RENDERER_2D_ENABLE_SAMPLE_FROM_RENDER_IMAGES`:
         *    -0. Image sampler for current render image (DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER)
         */
            /// @{
            /**
             * @brief Create descriptor pool and layout
             */
            void createDescriptorResources();
            /// @}

            // SWAPCHAIN RECREATION
            /**
             * @brief Swapchain recreation
             */
            /// @{
            /**
             * @brief Cleans up resources that were initialized by initSwapChainDependantResources()
             */
            void cleanupSwapChainDependantResources();
            /**
             * @brief Sets up resources that depend on the swap chain or its attributes
             * @details
             *  Initializes:
             *  - render pass
             *  - images, imageMemory, imageViews
             *  - image sampling descriptor set (if `RENDERER_2D_SAMPLE_FROM_RENDER_IMAGES`)
             *  - framebuffers
             *  - pipeline
             */
            void initSwapChainDependantResources();
            /**
             * @brief Recreates swap chain dependant resources
             * @details
             *  Calls:
             *  -# cleanupSwapChainDependantResources()
             *  -# initSwapChainDependantResources()
             */
            void swapChainRecreateCallback();
            /// @}

            Log rLog;
    };
}  // namespace gz::vlk
