#include "texture_manager.hpp"
#include "exceptions.hpp"
#include "vulkan_instance.hpp"

#include <glm/glm.hpp>

namespace gz::vlk {
    TextureManager::TextureManager(VulkanInstance& instance)
        : vk(instance)
    {
        LogCreateInfo logCI{
        	.logfile = "textureManager.log",
        	.storeLog = false,
        	.prefix = "Texture",
        	.prefixColor = Color::GREEN,
        	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
        };
        tLog = Log(std::move(logCI));

        vk.registerCleanupCallback(std::bind(&TextureManager::cleanup, this));
        atlases.emplace_back(TextureAtlas(vk, 24, 24, 24, 24));
        createDescriptorResources();
        VulkanInstance::registerObjectUsingVulkan("TextureManager",
                &descriptorSetLayout, &descriptorPool, &descriptorSet);
    }


    void TextureManager::cleanup() {
        tLog.log0("cleanup: cleaning up");
        vk.getDevice().destroyDescriptorSetLayout(descriptorSetLayout, NO_ALLOC);
        vk.getDevice().destroyDescriptorPool(descriptorPool, NO_ALLOC);
        for (auto& atlas : atlases) {
            atlas.cleanup();
        }
    }


	void TextureManager::getTexCoords(const std::string& textureName, glm::vec2& texCoords) {
        if (!textureInfos.contains(textureName)) {
            loadTextureFromFile(textureName);
        }
        TextureInfo& texI = textureInfos[textureName];
        tLog.log0("getTexCoords", texCoords, "textureInfo: topleft", texI.texCoordTopLeft, "botright", texI.texCoordBottomRight);
        texCoords = texI.texCoordTopLeft + texCoords * (texI.texCoordBottomRight - texI.texCoordTopLeft);
	}


	void TextureManager::loadTextureFromFile(const std::string& textureName) {
        int textureWidth, textureHeight, textureChannels;
        stbi_uc* pixels = stbi_load(("textures/" + textureName).c_str(), &textureWidth, &textureHeight, &textureChannels, STBI_rgb_alpha);
        if (!pixels) {
            throw Exception("Failed to load texture: textures/" + textureName, "TextureManager::loadTexture");
        }
        TextureInfo& texI = textureInfos[textureName];
        texI.atlas = 0;
        // TODO
        auto texCoords = atlases.at(0).addTexture(pixels, textureWidth, textureHeight);
        stbi_image_free(pixels);
        tLog.log0("TextureManager::loadTextureFromFile: TexCoords of new image:", texCoords.first, texCoords.second);
        texI.texCoordTopLeft = std::move(texCoords.first),
        texI.texCoordBottomRight = std::move(texCoords.second),
        tLog.log0("TextureManager::loadTextureFromFile: After loading:", atlases.at(0));
	}


//
// DESCRIPTOR SET
//
    void TextureManager::createDescriptorResources() {
        // 1) layout
        // combined image sampler
        std::vector<vk::DescriptorSetLayoutBinding> samplerLayoutBindings(atlases.size(), {
                .binding = 0,
                .descriptorType = vk::DescriptorType::eCombinedImageSampler,
                .descriptorCount = 1,
                .stageFlags = vk::ShaderStageFlagBits::eFragment,
                /* .pImmutableSamplers = nullptr, */
            }
        );
        vk.createDescriptorSetLayout(samplerLayoutBindings, descriptorSetLayout);

        // 2) pool
        vk::DescriptorPoolSize poolSize { 
            .type = vk::DescriptorType::eCombinedImageSampler,
            .descriptorCount = vk.getMaxFramesInFlight(),
        };
        vk.createDescriptorPool({ poolSize }, vk.getMaxFramesInFlight(), descriptorPool);

        // 3) Set
        vk.createDescriptorSet(descriptorSetLayout, descriptorPool, descriptorSet);

        // 4) configure sets
        std::vector<vk::DescriptorImageInfo> imageInfos;
        for (auto it = atlases.cbegin(); it != atlases.cend(); it++) {
            imageInfos.emplace_back();
            imageInfos.back().imageLayout = vk::ImageLayout::eShaderReadOnlyOptimal;
            imageInfos.back().imageView = it->getTextureImageView();
            imageInfos.back().sampler = it->getTextureSampler();
        }

        vk::WriteDescriptorSet descriptorW {
            .dstSet = descriptorSet,
            .dstBinding = 0,
            .dstArrayElement = 0,
            .descriptorCount = 1,
            .descriptorType = vk::DescriptorType::eCombinedImageSampler,
        };
        descriptorW.setImageInfo(imageInfos);
        
        // 1 write, no copies
        vk.getDevice().updateDescriptorSets(descriptorW, nullptr);
        tLog.log0("createDescriptorSets: Created descriptor sets.");
    }

}  // namespace gz::vlk
