#pragma once


#include "vulkan_util.hpp"

#include "dynamic_size_block_allocator.hpp"

#define LOG_SUBLOGS
#include <gz-util/log.hpp>

#include <initializer_list>

#define VULKAN_HPP_NO_EXCEPTIONS
#define VULKAN_HPP_NO_CONSTRUCTORS
#include <vulkan/vulkan.hpp>

#include <list>
#include <map>


namespace gz::vlk {
	/**
	 * @brief Contains information about a subsection (block) of a vk::DeviceMemory
	 */
	struct MemoryInfo {
		/**
		 * @brief Handle of the memory
		 */
		vk::DeviceMemory memory = VK_NULL_HANDLE;
		/**
		 * @brief Offset into memory
		 */
		vk::DeviceSize offset = 0;
		/**
		 * @brief The memoryTypeIndex memory was allocated from. Needed for VulkanAllocator::free()
		 */
		uint32_t memoryTypeIndex = 0;
	};

	/**
	 * @brief Manage a single vk::DeviceMemory chunk
	 */
	struct DeviceMemory {
		DeviceMemory() = delete;
		DeviceMemory(vk::DeviceSize size_);

		vk::DeviceSize size;
		vk::DeviceMemory memory;

		DynBlockAllocator dynBlockAllocator;
	};

	
	// if no memory is available, allocate a chunk of multiplier * requested size
	constexpr vk::DeviceSize TODO_ALLOCATION_SIZE_MULTIPLIIER = 10;
	

	// defined in vulkan_instance.hpp
	class VulkanInstance;

	/**
	 * @brief Allocator for device local memory
	 * @details
	 *  Allocates larger chunks of vk::DeviceMemory from which blocks can be @ref allocate() "allocated".
	 *
	 *  This class is for device local memory only, not for host memory. 
	 *  You can not use it for vulkan memory allocation callbacks (pAllocator).
	 *
	 *  @subsection usage How to use
	 *   The usage is explained with the example of creating and destroying a single buffer.
	 *   @code
	 *    // Create buffer handle
	 *    vk::Buffer buffer;
	 *    // Create empty memory info struct
	 *    MemoryInfo bufferMI;
	 *
	 *    vk::BufferCreateInfo bufferCI{}
	 *    ...
	 *    vk::createBuffer(...);
	 *
	 *	  // get memory requirements
     *    vk::MemoryRequirements2 memReq;
     *    vk::BufferMemoryRequirementsInfo2 bufMemReq;
	 *    // from vulkan_util.hpp, same exists for imageMemoryRequirements
     *    getBufferMemoryRequirements(device, buffer, bufMemReq, memReq);
	 *
	 *    // set allocation info
     *    vk::MemoryAllocateInfo memAI {
	 *    ...
     *    	.memoryTypeIndex = VulkanInstance::findMemoryType(memReq.memoryRequirements.memoryTypeBits, wantedProperties),
	 *
	 *	  // allocate memory for the buffer
     *    VulkanAllocator::allocate(memAI, memReq, bufferMI);
	 *
	 *	  // bind the buffer to the memory
     *    device.bindBufferMemory(buffer, bufferMI.memory, bufferMI.offset);
	 *
	 *    ...
	 *
	 *    device.destroyBuffer(buffer);
	 *    VulkanAllocator::free(bufferMemoryInfo);
	 *   @endcode
	 *
	 *   You will of course have to use a VulkanAllocator and VulkanInstance object, since the used methods are not static.
	 */
	class VulkanAllocator {
		public:
			VulkanAllocator(VulkanInstance& instance);
			/**
			 * @brief Get a block from a vk::DeviceMemory
			 * @details
			 *  Get a block of allocI.allocationSize of a vk::DeviceMemory that was allocated with allocI.memoryTypeIndex.
			 *  
			 *  When this function returns, memoryInfo will contain the information about the available block.
			 *  You can then bind something of size allocI.allocationSize to memoryInfo.memory at offset memoryInfo.offset.
			 * @throws vk::Exception when a call to vkAllocateMemory is needed and fails
			 * @todo Determine the size of new allocations
			 * @todo alignment, maybe use vk::MemoryRequirements instead of vk::MemoryAllocateInfo?
			 * @todo maybe increase the block size of the allocated block so that the following blocks is already 16-aligned?
			 */
			void allocate(const vk::MemoryAllocateInfo& allocI, const vk::MemoryRequirements2& memReq, MemoryInfo& memoryInfo);
			/**
			 * @brief Free a block allocated with allocate()
			 *
			 *  When this function returns, memoryInfo will be reset to default values
			 * @throws vk::UserError if memoryInfo was not allocated from this allocator
			 */
			void free(MemoryInfo& memoryInfo);

			/**
			 * @brief Deallocate all memory
			 * @details
			 *  Prints warnings if blocks have not been freed.
			 *
             * @note This function has to be called manually (by VulkanInstance). This class does not register a cleanupCallback with the instance!
			 *
			 */
			void cleanup();
		private:
			/// Allocated memory for memoryIndexType
			std::map<uint32_t, std::vector<DeviceMemory>> memory;
			Log aLog;
			/// Needed for access to logical device
			VulkanInstance& vk;

	};  // class VulkanAllocator
}  // namespace gz::vlk
