#pragma once

#include <memory>
#define GLM_FORCE_DEPTH_ZERO_TO_ONE
/* #include <glm/glm.hpp> */

#define GLFW_INCLUDE_VULKAN
#include <GLFW/glfw3.h>

#define VULKAN_HPP_NO_EXCEPTIONS
#define VULKAN_HPP_NO_CONSTRUCTORS
#include <vulkan/vulkan.hpp>

#define LOG_SUBLOGS

#include "vertex.hpp"
#include "vk_convert.hpp"
#include "vulkan_util.hpp"

#include "vulkan_allocator.hpp"
#include "vulkan_settings.hpp"

#include <gz-util/log.hpp>
#include <gz-util/settings_manager.hpp>
#include <gz-util/string/utility.hpp>

#include <cstdint>
#include <vector>

namespace gz::vlk {
    constexpr vk::MemoryMapFlags NO_MEM_FLAGS;
    constexpr vk::AccessFlags NO_ACC_FLAGS;
    constexpr vk::AccessFlags2 NO_ACC_2_FLAGS;
    constexpr vk::CommandBufferResetFlags NO_CMD_RESET_FLAGS;
    const uint32_t NO_OFFSET = 0;
    constexpr vk::AllocationCallbacks* NO_ALLOC = nullptr;
    /* const vk::ImageSubresourceRange DEFAULT_SUBRESOURCE_RANGE { */
    /*             .aspectMask = vk::ImageAspectFlagBits::eColor, */
    /*             .baseMipLevel = 0, */
    /*             .levelCount = 1, */
    /*             .baseArrayLayer = 0, */
    /*             .layerCount = 1, */
    /* }; */

    enum InstanceCommandPool {
        POOL_GRAPHICS, POOL_TRANSFER
    };

    /**
     * @nosubgrouping
     * @details
     *  @section vi_options
     *   - `VULKAN_INSTANCE_SWAP_CHAIN_AS_TRANSFER_DST`
     *   - `VULKAN_INSTANCE_SWAP_CHAIN_AS_RENDER_TARGET`
     *
     * @todo Push constancts for texture atlases
     * @todo Multiple texture atlases test
     * @todo Font
     * @todo Runtime glsl -> spirv compilation -> pipeline recreation
     * @todo Improve current allocation "strategy"
     */
    class VulkanInstance {
        public:
            /**
             * @name Public Interface: For the "creator" of the instance
             * @details
             *  These functions will probably called from your `main` function and loop.
             */
            /// @{
            VulkanInstance(gz::SettingsManagerCreateInfo<VULKAN_SETTINGS_MAN_TYPES>smCI) : settings(smCI), allocator(*this) {};
            /**
             * @brief Initializes the vulkan instance
             * @details
             *  -# @ref createWindow() "create a window through glfw"
             *  -# @ref createInstance() "create the vulkan instance"
             *  -# @ref setupDebugMessenger() "sets up the debug messenger"
             *  -# @ref createSurface() "create a the window surface"
             *  -# @ref selectPhysicalDevice() "select a GPU"
             *  -# @ref setValidSettings() "set the possible settings for the SettingsManager"
             *  -# @ref createLogicalDevice() "create a logical device"
             *  -# @ref createSwapChain() "create the swap chain"
             *  -# @ref createCommandPools() "create the command pools"
             *  @todo move depth image, texture, texture samples, model stuff to renderers
             *  -# @ref createCommandBuffers() "create command buffers for swap chain image layout transitions"
             *  -# @ref createSyncObjects() "create synchronization objects"
             */
            void init();
            /**
             * @brief Acquires a new frame from the swap chain and sets the image layout
             * @details
             *  How to draw stuff right now:
             *  -# Call this function
             *  -# Call the draw functions from the renderers.
             *     The renderers need to put the recorded command buffers into the commandBuffersToSubmitThisFrame vector using submitThisFrame()
             *  -# Call endFrameDraw
             */
            uint32_t beginFrameDraw();
            void endFrameDraw(uint32_t imageIndex);
            /**
             * @brief Cleanup the instance
             * @details
             *  -# wait for the device to idle
             *  -# call @ref registerCleanupCallback() "cleanup callbacks" in reverse order to the registration
             *  -# destroy command buffers
             *  -# call cleanupSwapChain()
             *  -# destroy @ref createSyncObjects() "synchronization objects"
             *  -# destroy @ref createCommandPools() "command pools"
             *  -# call VulkanAllocator::cleanup()
             *  -# destroy device
             *  -# destroy surface
             *  -# call cleanupDebugMessenger()
             *  -# destroy instance
             *  -# call glfwDestroyWindow()
             *  -# call glfwTerminate()
             */ 
            void cleanup();
            GLFWwindow* getWindow() { return window; }
            /// @}

        /**
         * @name Public Interface: init and cleanup
         * @details
         *  These functions can be used by other parts of your program that use this instance,
         *  like a renderer.
         */
        public:
            /// @{
            /**
             * @brief Destroys everything that was initalized in init 
             * @details
             *  -# Calls every callback registered by registerCleanupCallback() (FILO order)
             *  -# calls cleanupSwapChain()
             *  -# destroys everything initalized in init()
             *  -# destroys the window
             *  -# calls glfwTerminate()
             */
            void registerCleanupCallback(std::function<void()> callbackF);

            void registerSwapChainRecreateCallback(std::function<void()> callbackF);

            /**
             * @brief Register vulkan handles used by another object
             * @details
             *  Use this function to register vulkan handles used by another object. 
             *  When a @ref debugLog "validation message" contains one of these handles, the owner object can
             *  be determined and added to the message.
             *  This makes the debugging process easier when multiple objects use the same type of vulkan handle.
             *
             *  This function is a ObjectsUsingVulkan factory and does not take an ObjectsUsingVulkan ptr.
             *  The reason is that this makes it easy to make the function a nop when not in debug mode (meaning `NDEBUG` is defined).
             *  The entire call (and therefore construction of the ObjectUsingVulkan object) can then be optimized away.
             */
            template<PtrToVulkanHppHandleConvertible... T>
            static void registerObjectUsingVulkan(std::string&& name, const T&&... pHandleConvertibles);
            /// @}

        /**
         * @name Public Interface: Various utility
         * @details
         *  These functions can be used by other parts of your program that use this instance,
         *  like a renderer.
         */
        public:
            /// @{
            const vk::Device& getDevice() const { return device; }
			const vk::Extent2D& getScExtent() const { return scExtent; }
			const std::vector<vk::Image>& getScImages() const { return scImages; }
			const vk::Format& getScImageFormat() const { return scImageFormat; }
            const vk::Format& getDepthFormat() const { return depthFormat; }

            SettingsManager<VULKAN_SETTINGS_MAN_TYPES>& getSettings() { return settings; }

            /**
             * @brief Get the maximum number of frames in flight
             * @see MAX_FRAMES_IN_FLIGHT
             */
            uint32_t getMaxFramesInFlight() const { return MAX_FRAMES_IN_FLIGHT; }
            /**
             * @brief Get the frame that is currently in flight
             * @see currentFrame
             */
            uint32_t getCurrentFrame() const { return currentFrame; }

            void submitThisFrame(vk::CommandBuffer& cmdBuffer);

            /**
             * @brief Copy from srcBuffer to dstBuffer
             * @details
             *  Uses @ref beginSingleTimeCommands() "a single time command buffer" from commandPoolTransfer.
             */
            void copyBuffer(vk::Buffer srcBuffer, vk::Buffer dstBuffer, vk::DeviceSize size);

            /**
             * @brief Copy srcData to dstBuffer at offset via a staging buffer
             * @todo Detect when staging buffer is not necessary and make a direct copy
             */
            void copyToDeviceBuffer(const void* srcData, vk::DeviceSize srcSize, vk::Buffer& dstBuffer, MemoryInfo& dstMemory, uint32_t dstOffset);

            /**
             * @brief Begin a command buffer that is going to be used once
             * @param commandPool: The command pool from which the buffer should be allocated
             */
            vk::CommandBuffer beginSingleTimeCommands(InstanceCommandPool commandPool);
            /**
             * @brief Submit cmdBuffer on the queue and free it afterwards
             * @param cmdBuffer: Command buffer returned by beginSingleTimeCommands()
             * @param commandPool: The same pool as passed to beginSingleTimeCommands()
             * @details
             *  After command buffer is submitted on the specified queue, the function waits for the queue to idle
             */
            void endSingleTimeCommands(vk::CommandBuffer cmdBuffer, InstanceCommandPool commandPool);
            /// @}

        /**
         * @name Public Interface: Vulkan object creation/descruction
         * @details
         *  These functions aim at making creation and desctruction of common vulkan objects easier
         *  and can be used by other parts of your program that use this instance,
         *  like a renderer.
         */
        public:
            /// @{
            /**
             * @brief Create MAX_FRAMES_IN_FLIGHT command buffers from the commandPoolGraphics
             */
            void createCommandBuffers(std::vector<vk::CommandBuffer>& commandBuffers);
            /**
             * @brief Destroy all command buffers (must be from commandPoolGraphics)
             */
            void destroyCommandBuffers(std::vector<vk::CommandBuffer>& commandBuffers);

            /**
             * @brief Create a new graphics pipeline
             * @params pipelineCI @ref create_pipeline_create_info "(Not necessary fully initalized)" create info struct. 
             * @param desciptorSetLayouts [Optional] Vector of descriptor set layouts
             * @param useDepthStencil Wether to use depth
             * @param pipeline Pipeline handle to bind
             * @param T the type of Vertex to use, either Vertex2D or Vertex3D
             *
             * @details
             *  @section create_pipeline_ci Pipeline Create Info
             *   @subsection create_pipeline_ci_optional Optional and default values
             *    The pipelineCI struct does not have to be fully initalized.
             *    Some values can be `nullptr`, in which case a default value is used to create the pipeline.
             *    The optional values and their defaults are:
             *    - pVertexInputState: initalized from `VertexT::getBindingDescription()` and `VertexT::getAttributeDescriptions()`
             *    - pViewportState: viewport of swap chain extent and scissor of swap chain extent
             *    - pRasterizationState:
             *      - fill polygons 
             *      - backface culling 
             *      - counter clockwise front face
             *    - pMultisampleState: one sample
             *    - pColorBlendState: one rgba color blend attachment with 1-alpha color blend factor
             *    - pDepthStencilState: compare less, depth test and write enabled (only used when useDepthStencil is true!)
             *  @subsection create_pipeline_ci_required Required values
             *   The following values have to be provided in pipelineCI. VkUserError is thrown when
             *   one of these is nullptr:
             *   - renderPass
             *   - pStages (and therefor stageCount, which is not checked)
             *   - subpass (creates no exception, since the default 0 might be the actual value!)
             *
             *  @throws VkUserError when @ref create_pipeline_ci_required "required value" is not provided
             *  @throws VkException when pipeline (layout) creation fails
             *
             * @todo push constants, dynamicStates
             * 
             */
            template<VertexType T>
            void createGraphicsPipeline(vk::GraphicsPipelineCreateInfo&& pipelineCI, const std::vector<vk::DescriptorSetLayout>& descriptorSetLayouts, const std::vector<vk::PushConstantRange>& pushConstantRanges, bool useDepthStencil, Pipeline& pipeline);

            vk::ShaderModule createShaderModule(const std::vector<char>& code);
            /* void createShaderStageCreateInfos(const std::vector<std::pair<std::string, vk::SpecializationInfo*>>& shaders, std::vector<vk::PipelineShaderStageCreateInfo>& shaderStages); */

            /**
             * @brief Allocate memory, create a buffer and bind buffer to memory
             * @param size The size of the buffer
             * @param buffer The (null) handle to the buffer
             * @param bufferMemory Reference to an (uninitialized) MemoryInfo struct. It will be valid when the function returns
             * @param properties Properties that the buffers memory needs to satisfy
             * @param qFamiliesWithAccess [Optional] vector of queue families that have access to the buffer. Only needed with SHARING_MODE_CONCURRENT
             * @details
             *  The memory the buffer will be bound to is HOST_VISIBLE and HOST_COHERENT
             *
             */
            void createBuffer(vk::DeviceSize size, vk::BufferUsageFlags usage, vk::MemoryPropertyFlags properties, vk::Buffer& buffer, MemoryInfo& bufferMemory, vk::SharingMode sharingMode=vk::SharingMode::eExclusive, std::vector<uint32_t>* qFamiliesWithAccess=nullptr);
            /**
             * @brief Destroy buffer and free memory
             */
            void destroyBuffer(vk::Buffer& buffer, MemoryInfo& bufferMemory);

            /**
             * @brief Create a vertex buffer
             * @param vertexCount Number of vertices the buffer should contain
             * @param vertexBuffer The (null) handle to the buffer
             * @param vertexBufferMemory Reference to an (uninitialized) MemoryInfo struct. It will be valid when the function returns
             * @details
             *  The function is instantiated for Vertex2D and Vertex3D
             */
            template<VertexType VertexT>
            void createVertexBuffer(size_t vertexCount, vk::Buffer& vertexBuffer, MemoryInfo& vertexBufferMemory, vk::DeviceSize& vertexBufferSize);

            /**
             * @brief Create an index buffer
             * @param indexCount Number of indices the buffer should contain
             * @param indexBuffer The (null) handle to the buffer
             * @param indexBufferMemory Reference to an (uninitialized) MemoryInfo struct. It will be valid when the function returns
             * @details
             *  The function is instantiated for uint16_t and uint32_t
             */
            template<SupportedIndexType T>
            void createIndexBuffer(size_t indexCount, vk::Buffer& indexBuffer, MemoryInfo& indexBufferMemory, vk::DeviceSize& indexBufferSize);

            /**
             * @brief Create MAX_FRAMES_IN_FLIGHT uniform buffers
             */
            template<typename T>
            void createUniformBuffers(std::vector<vk::Buffer>& uniformBuffers, std::vector<MemoryInfo>& uniformBuffersMemory);

            /**
             * @brief Create as many framebuffers as images views, with the extent of the swap chain
             */
            void createFramebuffers(std::vector<vk::Framebuffer>& framebuffers, std::vector<vk::ImageView>& imageViews, vk::RenderPass& renderPass, vk::ImageView depthImageView=VK_NULL_HANDLE);
            /**
             * @brief Destroy the framebuffers
             */
            void destroyFramebuffers(std::vector<vk::Framebuffer>& framebuffers);

            /**
             * @brief Create a texture sampler
             * @todo add options and documentation
             */
            void createTextureSampler(vk::Sampler& sampler);

            void destroyTextureSampler(vk::Sampler& sampler);

            /**
             * @brief Create a descriptor layout with bindings
             */
            void createDescriptorSetLayout(const std::vector<vk::DescriptorSetLayoutBinding>& bindings, vk::DescriptorSetLayout& layout);
            void createDescriptorPool(const std::vector<vk::DescriptorPoolSize>& poolSizes, uint32_t maxSets, vk::DescriptorPool& pool);
            void createDescriptorSet(const vk::DescriptorSetLayout& layout, const vk::DescriptorPool& pool, vk::DescriptorSet& set);
            void createDescriptorSets(const std::vector<vk::DescriptorSetLayout>& layouts, const vk::DescriptorPool& pool, std::vector<vk::DescriptorSet>& sets);

        /**
         * @name Public Interface: Image utility
         * @details
         *  These functions can be used by other parts of your program that use this instance,
         *  like a renderer.
         */
            /// @{
            /**
             * @brief Create a 2D image, allocate the imageMemory and bind the image to it
             */
            void createImage(uint32_t width, uint32_t height, vk::Format format, vk::ImageTiling tiling, vk::ImageUsageFlags usage, vk::MemoryPropertyFlags memoryProperties, vk::Image& image, MemoryInfo& imageMI);
            void destroyImage(vk::Image& image, MemoryInfo& imageMemory);
            /**
             * @brief Create a 2D imageView with format for image.
             */
            void createImageView(vk::Format format, const vk::Image& image, vk::ImageView& imageView, vk::ImageAspectFlags aspectFlags);
            void destroyImageView(vk::ImageView& imageView);

            void copyBufferToImage(vk::Buffer buffer, vk::Image image, int32_t offsetX, int32_t offsetY, uint32_t width, uint32_t height);
            /**
             * @todo make a version using vkCmdResolveImage for multisampled images
             * @brief Copy srcImage to dstImage
             * @details
             *  Both images must have:
             *  - same extent
             *  - mipLevel 0
             *  - layerCount 1
             *  
             *  Calls vkCmdBlitImage, but does NOT submit the commandBuffer
             * @param cmdBuffer The command buffer where the command will be recorded to
             * @param srcImage Image with layout TRANSFER_SRC_OPTIMAL  
             * @param dstImage Image with layout TRANSFER_DST_OPTIMAL  
             */
            void copyImageToImage(vk::CommandBuffer& cmdBuffer, vk::Image srcImage, vk::Image dstImage, vk::Extent2D extent);  //, vk::CommandPool& commandPool=commandPoolTransfer);
            /**
             * @brief Transition the layout of image from oldLayout to newLayout
             * @details
             *  Supported transitions:
             *  - UNDEFINED -> DEPTH_STENCIL_ATTACHMENT_OPTIMAL (graphics q)
             *  - UNDEFINED -> TRANSFER_DST_OPTIMAL (transfer q)
             *  - UNDEFINED -> COLOR_ATTACHMENT_OPTIMAL (graphics q)
             *  - SHADER_READ_ONLY_OPTIMAL -> TRANSFER_DST_OPTIMAL (graphics q)
             *  - TRANSFER_DST_OPTIMAL -> SHADER_READ_ONLY_OPTIMAL (graphics q)
             *  - TRANSFER_DST_OPTIMAL -> PRESENT_SRC_KHR (graphics q)
             *  - COLOR_ATTACHMENT_OPTIMAL -> PRESENT_SRC_KHR (graphics q)
             *  - COLOR_ATTACHMENT_OPTIMAL -> TRANSFER_SRC_OPTIMAL (graphics q)
             *
             *  If you do not provide a command buffer, a command buffer from the indicated queue will be created and submitted.
             *  If you do provide a command buffer, the command is recorded but NOT submitted.
             *
             *  The transition is done by submitting a vk::ImageMemoryBarrier2
             *
             * @param cmdBuffer [Optional] The command buffer where the command will be recorded to
             * @throws vk::UserError if the transition is not supported.
             */
            void transitionImageLayout(vk::Image image, vk::Format format, vk::ImageLayout oldLayout, vk::ImageLayout newLayout, vk::CommandBuffer* cmdBuffer=nullptr);
            /// @}


            
        /**
         * @name Debugging (validation messages)
         */
        public:
            /// @{
            /**
             * @brief Log messages from validation layers with the Apps logger
             * @details
             *  Using the static vLog to log vulkan messages with a prefix dependant on the messageType.
             *  - Validation and performace messages usually have a bracket containing a VUID, 
             *    these brackets are printed in yellow/red when its a warning/error.
             *  - Handles printed by the validation layer are checked with objectsUsingVulkan, if found
             *    the owner of the handle is added to the message @see registerObjectUsingVulkan(), ObjectUsingVulkan
             *  - Different message sources (layer, shader, general...) get a prefix with a different color
             */
            static VKAPI_ATTR VkBool32 VKAPI_CALL debugLog(
                    VkDebugUtilsMessageSeverityFlagBitsEXT messageSeverety,
                    VkDebugUtilsMessageTypeFlagsEXT messageType,
                    const VkDebugUtilsMessengerCallbackDataEXT* pCallbackData,
                    void* pUserData);
            static Log createSublog(bool showLog, const std::string& prefix, Color prefixColor) {
                return vLog.createSublog(showLog, prefix, prefixColor);
            }
            /// @}

        private:
        /**
         * @name Settings
         */
            /// @{
            gz::SettingsManager<VULKAN_SETTINGS_MAN_TYPES> settings;
            /**
             * @brief Set valid values for the SettingsManager according to phDevFeatures and phDevProperties
             * @details
             *  Must be called after selectPhysicalDevice() and before createLogicalDevice()
             *  Sets valid values for:
             *  - anisotropy_enable (bool) [features.samplerAnisotropy]
             *  - max_anisotropy (float) [1-limits.maxSamplerAnisotropy]
             *  - max_frames_in_flight (uint32_t) [1-4]
             *
             */
            void setValidSettings();
            /**
             * @brief The maximum number of frames that can be ready for presentation
             * @details
             *  A higher number can lead to more stable frame rates, but also increases latency.
             *
             *  This variable is assigned from setting "max_frames_in_flight". 
             * @warning 
             *  Do not use the setting from the settings manager: 
             *  A lot of resources depend on this variable and if it gets changed at runtime, the program would crash.
             *  Always use this variable (or the @ref getMaxFramesInFlight() "getter").
             *  Changes to max_frames_in_flight will take effect upon restart.
             */
            uint32_t MAX_FRAMES_IN_FLIGHT;
            /// @}

        /**
         * @name Window
         */
            /// @{
            GLFWwindow* window;
            void createWindow();
            /// @}

        /**
         * @name Framebuffers
         */
            /// @{
            bool frameBufferResized = false;
            static void frameBufferResizedCallback(GLFWwindow* window, int width, int height);

            /**
             * @brief The frame that is currently in flight
             * @details
             *  The value is between 0 and MAX_FRAMES_IN_FLIGHT - 1.
             *  It specifies which command buffer or uniform buffer to use for drawing the current frame,
             *  it is not the image index for the current @ref scImages "swap chain image".
             */
            uint32_t currentFrame = 0;
            /// @}

        /**
         * @name Cleanup
         */
            /// @{
            std::vector<std::function<void()>> cleanupCallbacks;
            std::vector<std::function<void()>> scRecreateCallbacks;
            /// @}

        /**
         * @name Instance
         */
            /// @{
            vk::Instance instance;
            /**
             * @brief Create the vulkan instance
             * @details
             *  -# check if validationLayers are available (if enabled)
             *  -# create instance with info
             *  -# check if all extensions required by glfw are available
             * @throws VkException if:
             *  - validation layers are enabled but not supported 
             *  - instance creation fails
             *  - not all required instance extensions are available
             */
            void createInstance();
            /**
             * @brief Get all the required instance extensions
             * @details
             *  The required extensions are:
             *  - all extensions required by glfw (glfwGetRequiredInstanceExtensions())
             *  - if validation layers are enabled: "VK_EXT_debug_utils"
             */
            std::vector<const char*> getRequiredInstanceExtensions() const;
            /**
             * @brief Check if @ref settings::validationLayers "all requrested validation layers" are supported
             */
            bool validationLayersSupported();
            /// @}

        /**
         * @name Physical device (selection process)
         * @brief Member variables and functions that are used to select a physical device
         */
            /// @{
            /**
             * @brief Assign the physicalDevice handle to the @ref rateDevice "best rated" GPU
             * @details
             *  After this method, the following members will be initalized:
             *  - physicalDevice
             *  - phDevProperties 
             *  - phDevFeatures
             *  - phDevMemProperties
             *  - qFamilyIndices
             *  - depthFormat
             */
            void selectPhysicalDevice();
            /**
             * @brief Assigns a score to a physical device.
             * @details
             *  score = GPU_TYPE_FACTOR + VRAM_SIZE (in MB) + SUPPORTED_QUEUES_FACTOR + FEATURES_FACTORS
             *  A GPU is unsuitable and gets score 0 if it does not:
             *  - have the needed queue families
             *  - support all necessary extensions
             *  - have swap chain support
             */
            unsigned int rateDevice(vk::PhysicalDevice device);
            /**
             * @brief Get the required physical device features
             * @details
             *  The required features are:
             *  - synchronization2
             */
            PhysicalDeviceFeatures getRequiredDeviceFeatures() const;

        /**
         * @name Physical device (selected)
         * @brief Member variables and functions that are usable/initalized after selectPhysicalDevice
         */
            /// @{
            vk::PhysicalDevice physicalDevice = VK_NULL_HANDLE;
            /// all the properties of the selected physcial device
            vk::PhysicalDeviceProperties2 phDevProperties;
            /// memory properties of the selected physical device
            vk::PhysicalDeviceMemoryProperties2 phDevMemProperties;
            /// all the features that the selected physical device supports
            PhysicalDeviceFeatures phDevFeatures;
            /// preferred depth format
            vk::Format depthFormat;

            /**
             * @brief Find the best of the supported formats
             * @param candidates Candidate format, from best to worst
             * @returns The first format from candidates that is supported
             * @throws VkException if no suitable format is found
             */
            vk::Format findSupportedFormat(const std::vector<vk::Format>& candidates, vk::ImageTiling tiling, vk::FormatFeatureFlags features);
            /// @}

        /**
         * @name Memory management
         */
            /// @{
            uint32_t findMemoryType(uint32_t typeFilter, vk::MemoryPropertyFlags properties);
            VulkanAllocator allocator;
            /// @}

        /**
         * @name Queue
         */
            /// @{
            QueueFamilyIndices findQueueFamilies(vk::PhysicalDevice device);
            QueueFamilyIndices qFamilyIndices;
            vk::Queue graphicsQ;
            vk::Queue presentQ;
            // if possible, use dedicated transferQ. if not available, transferQ = graphicsQ
            vk::Queue transferQ;
            /// @}

        /**
         * @name Logical device
         */
            /// @{
            vk::Device device;
            /**
             * @details
             *  request anisotropic sampling feature
             */
            void createLogicalDevice();
            /// @}

        /**
         * @name Surface
         */
            /// @{
            vk::SurfaceKHR surface;
            void createSurface();
            /// @}

        /**
         * @name Swap chain
         */
            /// @{
            vk::SwapchainKHR swapChain;
            std::vector<vk::Image> scImages;
            std::vector<vk::ImageView> scImageViews;
            vk::Format scImageFormat;
            vk::Extent2D scExtent;

            SwapChainSupport querySwapChainSupport(vk::PhysicalDevice device);
            /**
             * @todo Rate formats if preferred is not available
             * @todo Move back to SurfaceFormatKHR? 2 is an (for now) unnecessary extension
             */
            vk::SurfaceFormat2KHR selectSwapChainSurfaceFormat(const std::vector<vk::SurfaceFormat2KHR>& availableFormats);
            /**
             * @todo Check settings for preferred mode
             */
            vk::PresentModeKHR selectSwapChainPresentMode(const std::vector<vk::PresentModeKHR>& availableModes);
            vk::Extent2D selectSwapExtent(const vk::SurfaceCapabilities2KHR& capabilities);
            /**
             * @brief Create swap chain, image views
             * @todo query if transfer dst is supported
             */
            void createSwapChain();
            /**
             * @brief Destroy swap chain, image views
             */
            void cleanupSwapChain();
            /**
             * @brief Recreate the swap chain for the current window size
             * @details
             *  Does:
             *  -# get new window dimensions
             *      - blocks while dimensions == 0 (minimized)
             *  -# calls cleanupSwapChain
             *  -# calls createSwapChain
             *  -# other stuff
             *  -# calls all callbacks registered with registerSwapChainRecreateCallback
             */
            void recreateSwapChain();
            /// @}

        /**
         * @name Command pools
         */
            /// @{
            vk::CommandPool commandPoolGraphics;
            vk::CommandPool commandPoolTransfer;
            void createCommandPools();
            /// @}
        /**
         * @name Command buffers
         */
            /// @{
            std::vector<vk::CommandBuffer> commandBuffersBegin;
            std::vector<vk::CommandBuffer> commandBuffersEnd;
            /**
             * @brief All command buffers to submit on graphics queue for the current frame
             * @details
             *  Use submitThisFrame() to submit a command buffer
             */
            std::vector<vk::CommandBuffer> commandBuffersToSubmitThisFrame;
            /// @}

        /** 
         * @name Synchronization
         */
            /// @{
            std::vector<vk::Semaphore> imageAvailableSemaphores;
            std::vector<vk::Semaphore> renderFinishedSemaphores;
            std::vector<vk::Fence> inFlightFences;
            void createSyncObjects();
            /// @}

        /**
         * @name Utility
         */
            /// @{
            template<typename T, typename... Args>
            vk::Result runVkResultFunction(const char* name, Args&&... args);
            template<typename T, typename... Args>
            void runVkVoidFunction(const char* name, Args&&... args);
 
        /**
         * @name Debugging: (log and validation debug messenger)
         */
            /// @{
            void setupDebugMessenger();
            void cleanupDebugMessenger();
            VkDebugUtilsMessengerEXT debugMessenger;
            static gz::Log vLog;
            /// @}

        /**
         * @name Debugging: (handle ownership)
         */
            /// @{
            static std::vector<std::unique_ptr<ObjectUsingVulkanBase>> objectsUsingVulkan;
            /**
             * @brief Search for handles in message and get their owners
             * @details
             *  Searches for handles (hex numbers) in message and sets handleOwnerString.\n
             *  If message contains no handle, handleOwnerString will be empty. 
             *  If message contains at least one handle, handleOwnerString will be:
             *  `Handle Ownerships: [ <handle> - <owner>, ...]`
             */
            static void getHandleOwnerString(std::string_view message);
            /// Temporary used in getHandleOwnerString
            static std::string handleOwnerString;
            /// Temporary used in getHandleOwnerString to filter duplicates
            static std::set<uint64_t> foundHandles;
            /// Used for alternating the color of validation messages between grey and white
            static bool lastColorWhite;
            /// @}

    }; // VulkanInstance

//
// IMPLEMENTATIONS
//
// PUBLIC INTERFACE: INIT AND CLEANUP
    template<PtrToVulkanHppHandleConvertible... T>
    void VulkanInstance::registerObjectUsingVulkan(std::string&& name, const T&&... pHandleConvertibles) {
        objectsUsingVulkan.emplace_back(std::unique_ptr<ObjectUsingVulkanBase>(std::make_unique<ObjectUsingVulkan<T...>>(ObjectUsingVulkan(std::move(name), std::forward<const T>(pHandleConvertibles)...))));
    }

// PUBLIC INTERFACE: CREATION AND DESTRUCTION
    // UNIFORM BUFFERS
    template<typename T>
    void VulkanInstance::createUniformBuffers(std::vector<vk::Buffer>& uniformBuffers, std::vector<MemoryInfo>& uniformBuffersMemory) {
        vk::DeviceSize bufferSize = sizeof(T);

        uniformBuffers.resize(MAX_FRAMES_IN_FLIGHT);
        uniformBuffersMemory.resize(MAX_FRAMES_IN_FLIGHT);

        for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
            createBuffer(bufferSize, vk::BufferUsageFlagBits::eUniformBuffer, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, uniformBuffers[i], uniformBuffersMemory[i]);
        }
    }

// UTILITY
    template<typename T, typename... Args>
    vk::Result VulkanInstance::runVkResultFunction(const char* name, Args&&... args) {
        auto f = reinterpret_cast<T>(vkGetInstanceProcAddr(instance, name));
        if (f == nullptr) {
            vLog.error("getVkFunction: Could not get function:", name);
            throw std::runtime_error("getVkFunction: Could not get function.");
        }
        else {
            return vk::Result(f(std::forward<Args>(args)...));
        }
    };

    template<typename T, typename... Args>
    void VulkanInstance::runVkVoidFunction(const char* name, Args&&... args) {
        auto f = reinterpret_cast<T>(vkGetInstanceProcAddr(instance, name));
        if (f == nullptr) {
            vLog.error("getVkFunction: Could not get function:", name);
            throw std::runtime_error("getVkFunction: Could not get function.");
        }
        else {
            f(std::forward<Args>(args)...);
        }
    }

}
/**
 * @file Vulkan instance
 */

